#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <dirent.h>                     /* opendir, readdir */
#ifndef NOTIMEH
#include <time.h>                       /* time for seed */
#endif
#include <memory.h>                     /* memset() */
#include "hl.h"

extern hfield *hst;          /* stack of heightfields */
extern int hp;                 /* pointer to top of stack: hst[hp] */
extern int stacksize;          /* total size of stack */
extern int histbins;           /* histogram bins */
extern int takefile;           /* if a command file is being read in */
extern int tile_mode;          /* tiling mode: on/off/auto */
extern D tile_tol;             /* tiling edge difference threshold */

/* --- cplx.c functions -------------------------------------- */

int c_swap();                   /* swap complex and real parts of matrix */
int c_join();           /* make complex from real & imag parts of matrix */
int c_split();          /* divide complex into real & imag parts */
int c_real();           /* take real part of array */
int c_diff();           /* make complex from x & y slope of matrix */
int c_int();            /* generate real mx from complex x&y diff'l */
int c_mag();            /* take magnitude of complex matrix */
int c_convert(int dir); /* convert complex Polar<->Rect representation */

int c_swap()  /* swap complex and real parts of matrix */
{
int ix,iy;
int xsize, ysize;
PTYPE *hf;
hfield chf;
D tmp;


  if (s_empty()) return(1);     /* check if stack is empty */
  hf = hst[hp].a;
  xsize = hst[hp].xsize;
  ysize = hst[hp].ysize;

  if (hst[hp].c) {              /* swap real, imag parts of mx */
    for (ix=0;ix<xsize;ix++) {
      for (iy=0;iy<ysize;iy++) {
        tmp = El(hf,ix,iy);
        El(hf,ix,iy) = Im(hf,ix,iy);
        Im(hf,ix,iy) = tmp;
      }
    }
  } else {              /* matrix is real; must create new complex mx */

    if (h_newc(&chf,xsize,ysize,hst[hp].name)) return(1);

    for (ix=0;ix<xsize;ix++) {   /* old real part -> new imag. part */
      for (iy=0;iy<ysize;iy++) {
        Im(chf.a,ix,iy) = El(hf,ix,iy);
      }
    }
    free(hst[hp].a);   /* junk the old real matrix */
    hst[hp] = chf;   /* graft on the new complex matrix */

  }  /* end if cflag */
  h_minmax(); /* new real extrema */
  return(0);
} /* c_swap() */

int c_mag()  /* take magnitude of complex matrix */
{
int ix,iy;
int xsize, ysize;
hfield hfr,hfc;

  if (s_empty()) return(1);     /* check if stack is empty */
  hfc = hst[hp];
  if (!hfc.c) {
    printf("cmag error: matrix not complex.\n");
    return(1);
  }
  xsize = hfc.xsize;
  ysize = hfc.ysize;
  if (h_newr(&hfr,xsize,ysize,"cmag")) return(1);

  /* compute magnitude of matrix */

  for (ix=0;ix<xsize;ix++) {
      for (iy=0;iy<ysize;iy++) {
        El(hfr.a,ix,iy) = sqrt(pow(El(hfc.a,ix,iy),2)+pow(Im(hfc.a,ix,iy),2));
      }
  }

  free(hfc.a);   /* junk the old complex matrix */
  hst[hp] = hfr;   /* replace with the new real matrix */

  h_minmax(); /* new real extrema */
  return(0);
} /* c_mag() */


int c_convert(int dir)  /* convert complex matrix Rect <-> Polar */
{
int ix,iy;
int xsize, ysize;
hfield hfc;
D re, im;        /* temporary variables */

  if (s_empty()) return(1);     /* check if stack is empty */
  hfc = hst[hp];
  if (!hfc.c) {
    printf("convert error: matrix not complex.\n");
    return(1);
  }
  xsize = hfc.xsize;
  ysize = hfc.ysize;

  /* convert matrix */
  if (dir==0) {           /* rectangular -> polar */
    for (ix=0;ix<xsize;ix++) {
      for (iy=0;iy<ysize;iy++) {
        re = El(hfc.a,ix,iy);
        im = Im(hfc.a,ix,iy);
        El(hfc.a,ix,iy) = sqrt(re*re + im*im);
        Im(hfc.a,ix,iy) = ATAN2(im,re);
      }
    }
  } else {                /* polar -> rectangular */
    for (ix=0;ix<xsize;ix++) {
      for (iy=0;iy<ysize;iy++) {
        re = El(hfc.a,ix,iy);
        im = Im(hfc.a,ix,iy);
        El(hfc.a,ix,iy) = re * cos(im);
        Im(hfc.a,ix,iy) = re * sin(im);
      }
    }
  }

  h_minmax(); /* new real extrema */
  return(0);
} /* c_convert() */


int c_join()  /* make complex from real & imag parts of matrix */
{
int ix,iy;
int xsize, ysize;
hfield hfc,hfr,hfi;

  if (s_empty()) return(1);     /* check if stack is empty */
  hfr = hst[hp];
  xsize = hfr.xsize;
  ysize = hfr.ysize;
  DECSTACK;
  hfi = hst[hp];
  INCSTACK;
  if (hfi.a==NULL) { printf("c_join: two matrices needed\n"); return(1); }
  if (xsize != hfi.xsize || ysize != hfi.ysize) {
    printf("c_join: matrices must be of same size\n"); return(1); }

  if (h_newc(&hfc,xsize,ysize,"cjoin")) return(1);

  for (ix=0;ix<xsize;ix++) {   /* old real part -> new imag. part */
      for (iy=0;iy<ysize;iy++) {
        El(hfc.a,ix,iy) = El(hfr.a,ix,iy);
        Im(hfc.a,ix,iy) = El(hfi.a,ix,iy);
      }
  }
  
  hpop();   /* trash old matrices */
  hpop();

  INCSTACK;
  if (hst[hp].a!=NULL) free(hst[hp].a);  /* get rid of old matrix here */
  hst[hp] = hfc;
  
  h_minmax(); /* new real extrema */
  return(0);
} /* c_join */


int c_split()  /* divide complex into real & imag parts */
{
int ix,iy;
int xsize, ysize;
hfield hfr, hfi, hfc;

  if (s_empty()) return(1);     /* check if stack is empty */
  if (!hst[hp].c) {
    printf("csplit error: matrix not complex.\n");
    return(1);
  }
  hfc = hst[hp];
  xsize = hfc.xsize;
  ysize = hfc.ysize;
  if (h_newr(&hfr,xsize,ysize,"real")) return(1);
  if (h_newr(&hfi,xsize,ysize,"imag")) return(1);

  for (ix=0;ix<xsize;ix++) { 
      for (iy=0;iy<ysize;iy++) {
        El(hfr.a,ix,iy) = El(hfc.a,ix,iy);
        El(hfi.a,ix,iy) = Im(hfc.a,ix,iy);
      }
  }
  hpop();   /* trash old complex matrix */

  INCSTACK;
  if (hst[hp].a!=NULL) free(hst[hp].a);  /* get rid of old stack element */
  hst[hp] = hfi;
  h_minmax(); /* new extrema */

  INCSTACK;
  if (hst[hp].a!=NULL) free(hst[hp].a);  /* get rid of old stack element  */
  hst[hp] = hfr;
  h_minmax(); /* new extrema */

  return(0);
} /* c_split */

#ifdef BAD_REALLOC
int c_real()  /* take real part of array */
{
int ix,iy;
int xsize, ysize;
hfield h0,h1;

  if (s_empty()) return(1);     /* check if stack is empty */
  h0 = hst[hp];
  if (!h0.c) return(0);    /* nothing to do */

  xsize = h0.xsize;
  ysize = h0.ysize;
  if (h_newr(&h1,xsize,ysize,h0.name)) return(1);
  for (iy=0;iy<ysize;iy++) {
   for (ix=0;ix<xsize;ix++) {
     El(h1.a,ix,iy) = El(h0.a,ix,iy);
   }
  }
  h1.min = h0.min;
  h1.max = h0.max;
  free(h0.a);
  hst[hp] = h1;
  return(0);
} /* c_real */
#else
int c_real()  /* take real part of array */
{
int xsize, ysize;
hfield hf;
size_t memsize;

  if (s_empty()) return(1);     /* check if stack is empty */
  hf = hst[hp];
  if (!hf.c) return(0);    /* nothing to do */

  xsize = hf.xsize;
  ysize = hf.ysize;
  
  memsize = (size_t)xsize*ysize*sizeof(PTYPE);

                     /* hf.a assignment fixed by R.Gardener 4/26/07 */
  hf.a = (PTYPE*)realloc(hf.a,memsize);   /* truncate array to real */
  hf.c = FALSE;

  hst[hp] = hf;
  return(0);
} /* c_real */
#endif

int c_diff()  /* make complex from x & y slope of matrix */
{
int ix,iy;
int xsize, ysize;
hfield hf, hfc;
int tile;
D xdiff, ydiff, tmp;

  if (s_empty()) return(1);     /* check if stack is empty */
  hf = hst[hp];
  xsize = hf.xsize;
  ysize = hf.ysize;
  tile = h_tilable(hf,0);

  if (h_newc(&hfc,xsize,ysize,"cdiff")) return(1);

  for (ix=0;ix<xsize;ix++) {   /* old real part -> new imag. part */
      for (iy=0;iy<ysize;iy++) {
        tmp = El(hf.a,ix,iy);
        if (tile) {
          xdiff = tmp-Elmod(hf.a,ix-1,iy);
          ydiff = tmp-Elmod(hf.a,ix,iy-1);
        } else {
          xdiff = tmp-Elclip(hf.a,ix-1,iy);
          ydiff = tmp-Elclip(hf.a,ix,iy-1);
        }
        El(hfc.a,ix,iy) = xdiff;
        Im(hfc.a,ix,iy) = ydiff;
      }
  }
  
  hpop();   /* trash old matrix */

  INCSTACK;
  if (hst[hp].a!=NULL) free(hst[hp].a);  /* get rid of old matrix here */
  hst[hp] = hfc;
  h_minmax(); /* new real extrema */

  return(0);
} /* c_diff */

int c_int()
{
 int xsize,ysize;
 int ix,iy;
 hfield hf0,hf1;

  if (s_empty()) return(1);     /* check if stack is empty */
  hf0 = hst[hp];
  if (!hf0.c) {
   printf("cint error: input must be complex.\n");
   return(1);
  }
  xsize = hf0.xsize;
  ysize = hf0.ysize;
  if (h_newr(&hf1,xsize,ysize,"cint")) return(1);
  El(hf1.a,0,0) = 0;        /* start with corner [0,0] = 0.0 */
  for (ix=1;ix<xsize;ix++)
    El(hf1.a,ix,0) = El(hf1.a,ix-1,0) + El(hf0.a,ix,0);
  for (iy=1;iy<ysize;iy++)
    El(hf1.a,0,iy) = El(hf1.a,0,iy-1) + Im(hf0.a,0,iy);

  for (ix=1;ix<xsize;ix++) {
    for (iy=1;iy<ysize;iy++) {
        El(hf1.a,ix,iy) = 0.5 * (El(hf1.a,ix-1,iy)+El(hf0.a,ix,iy) +
                                 El(hf1.a,ix,iy-1)+Im(hf0.a,ix,iy)   );
    }
  }

  free(hf0.a);      /* free old HF */
  hst[hp] = hf1;    /* replace with new HF */
  h_minmax();       /* find min,max values */

  return(0);
} /* c_int() */
