/* dirfun.c  --  functions to handle file-directory related things */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <dirent.h>                     /* opendir, readdir */
#include <unistd.h>                     /* unlink() */
#include <time.h>                       /* ctime() */
#include "hl.h"

int f_delete(char *fname)
{
   if(unlink(fname)) {perror(fname); return(1);}
   else return(0);
}

void listdir(char *dname)          /* print directory entries */
{
 int chars;
 int maxlen;            /* maximum length of a filename */
 struct dirent *de;
 DIR *d; 
 
 d = opendir(dname);      /* run through all entries to find longest one */
 maxlen = 0;
 while ((de = readdir(d)) != NULL) {
   if ((strlen(de->d_name) > maxlen) && (de->d_name[0] != '.'))
        maxlen = strlen(de->d_name);
 }
 closedir(d);
 maxlen += 2;                   /* add two buffer spaces for neat columns */
 if (maxlen > 79) maxlen = 79;  /* don't make the column > 80-char line */

 d = opendir(dname);
 chars = 0;
 while ((de = readdir(d)) != NULL) {
  if (de->d_name[0] != '.') {        /* skip entries with a leading '.' */
   printf("%s ",de->d_name);
   chars += strlen(de->d_name) + 1;
   if ((chars+maxlen) > 70) { 
     printf("\n");
     chars = 0;
   } else               /* fill out even # of spaces between listings */
     for (;(float)chars/maxlen!=chars/maxlen;printf(" "),chars++);
  }
 }
 closedir(d);
 if (chars != 0) printf("\n");
 printf("\n");
} /* listdir() */


void glistdir(char *dname) /* print only graphics file directory entries */
{
 int chars;
 int maxlen;            /* maximum length of a filename */
 struct dirent *de;
 DIR *d;

 d = opendir(dname);      /* run through all entries to find longest one */
 maxlen = 0;
 while ((de = readdir(d)) != NULL) {
   if ((strlen(de->d_name) > maxlen) && is_gfile(de->d_name))
        maxlen = strlen(de->d_name);
 }
 closedir(d);
 maxlen += 2;                   /* add two buffer spaces for neat columns */
 if (maxlen > 79) maxlen = 79;  /* don't make the column > 80-char line */

 d = opendir(dname);
 chars = 0;
 while ((de = readdir(d)) != NULL) {
  if (is_gfile(de->d_name)) {        /* skip non-heightfield entries  */
   printf("%s ",de->d_name);
   chars += strlen(de->d_name) + 1;
   if ((chars+maxlen) > 70) {
     printf("\n");
     chars = 0;
   } else               /* fill out even # of spaces between listings */
     for (;(float)chars/maxlen!=chars/maxlen;printf(" "),chars++);
  }
 }
 closedir(d);
 if (chars != 0) printf("\n");
 printf("\n");
} /* glistdir() */

int is_gfile(char *f)   /* returns 1 if file has known extension (eg .TGA) */
{
 char ext[5];
 int pos,i;

 pos = strlen(f)-4;
 if (pos<1) return(0);
 for (i=0;i<4;i++) ext[i]=f[pos+i];
 ext[4]='\0';

 if ( s_equal(ext,".TGA") ||
      s_equal(ext,".GIF") ||
      s_equal(ext,".POT") ||
      s_equal(ext,".MAT") ||
      s_equal(ext,".OCT") ||
      s_equal(ext,".PGM") ||
      s_equal(ext,".PNG")   ) return(1);
      else return(0);
} /* is_gfile() */
