/* interp.c -- call functions according to arguments in insp[][]
 * uses global vars:  hst[], hp
 * sets global vars:  cmd_stream, int_done
 *
 * returns 0 if operation sucessful, 1 if an error ocurred
 */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <dirent.h>                     /* opendir, readdir */
#include <unistd.h>                     /* unlink() */
#include <time.h>                       /* ctime() */
#include "hl.h"
#include "hf_io.h"

extern hfield *hst;         /* stack of heightfields */
extern int hp;              /* pointer to top of stack: hst[hp] */
extern int stacksize;       /* size of HF stack */
extern int int_done;        /* interp-loop-exit flag */
extern int takefile;        /* if a command file is being read in */
extern int startint;        /* if HF-Lab session started in interactive mode */
extern int view_current;    /* if currently in view mode */
extern char cmd_instr[CLINELEN]; /* command line string */
extern int cmd_p;           /* pointer into cmd_instr[] */

extern FILE *cmd_stream;       /* input stream (stdin or command file) */

int cmd_seq(int cmd_ptr);  /* interprets raw string of commands */
int cmd_seq1(char instr[CLINELEN]);  /* run a command sequence a;b;c etc */
int parse_1tok(char *instr, int *t_ptr, char *tok_s); /* parse 1 token */
int h_interp(int tokens, char insp[MAXTOK][TOKLEN]); /* interpret one cmd. */
int parse_line(char *instr, int *t_ptr, char insp[MAXTOK][TOKLEN]);
int cmd_rep(int tokens, char insp[MAXTOK][TOKLEN]); /* repeat commands */

int h_interp(int tokens, char insp[MAXTOK][TOKLEN]) /* interpret commands */
{    
 float r1,r2,r3,r4,r5;
 int r=0;
 int i;
 char buf[80];
 char logname[80];              /* logfile name */

    initgauss();	/* always update random number seed */

    if (insp[0][0]=='#') { 
      for (i=0;i<tokens;i++) printf("%s ",insp[i]);
      printf("\n");
    } else if (keymatch(insp[0], "help", 1) || keymatch(insp[0],"?",1)) {
      if (tokens>1) r=pr_help(insp[1]);
      else r=pr_help("");
    } else if (tokens>1 && keymatch(insp[1], "?", 1)) { /*  '<cmd> ?' */
      r=pr_help(insp[0]);
    } else if (keymatch(insp[0], "sample", 3)) { /* print one element */
      r = h_sample(tokens,insp);
    } else if (keymatch(insp[0], "print", 3)) { /* print hf onscreen */
      r = h_print(tokens,insp);
    } else if (keymatch(insp[0], "tick", 3)) {
      r = time_tic();
    } else if (keymatch(insp[0], "tock", 3)) {
      r = time_toc();
    } else if (keymatch(insp[0], "rescale", 3)) {
      if (tokens<2) {printf("rescale: must specify dimension\n"); return(1);}
      sscanf(insp[1], "%f", &r1);
      if (tokens > 2) sscanf(insp[2], "%f", &r2); else r2 = r1;
      r=rescale((int) r1,(int) r2);
    } else if (keymatch(insp[0], "clip", 4)) {
      if (tokens<3) 
	{printf("clip: must specify at least one point\n"); return(1);}
      sscanf(insp[1], "%f", &r1);
      if (tokens>2) sscanf(insp[2], "%f", &r2); else r2=r1;
      if (tokens>3) sscanf(insp[3], "%f", &r3); else r3=0;
      if (tokens>4) sscanf(insp[4], "%f", &r4); else r4=r3;
      r=h_clip((int)r3,(int)r4,(int)r1,(int)r2);
    } else if (keymatch(insp[0], "random",3)) {
      if ((tokens==1) && (hst[hp].a==NULL)) 
	{printf("must specify a dimension.\n"); return(1);}
      if (tokens==1) {r1 = hst[hp].xsize; r2 = hst[hp].ysize; }
      else { if (tokens > 1) sscanf(insp[1], "%f", &r1); 
	     if (tokens > 2) sscanf(insp[2], "%f", &r2); else r2 = r1;
	  }
      r=gen_rand((int) r1,(int) r2);
    } else if (keymatch(insp[0], "zero",4)) { 
      if ((tokens==1) && (hst[hp].a==NULL)) 
	{printf("must specify a dimension.\n"); return(1);}
      if (tokens==1) {r1 = hst[hp].xsize; r2 = hst[hp].ysize; }
      else { if (tokens > 1) sscanf(insp[1], "%f", &r1); 
	     if (tokens > 2) sscanf(insp[2], "%f", &r2); else r2 = r1;
	  }
      r=zero((int)r1,(int)r2);        /* create a new blank HF */
    } else if (keymatch(insp[0], "const",5)) { 
      r3 = 0;
      if ((tokens<3) && (hst[hp].a==NULL)) 
	{printf("stack empty: must specify dimension.\n"); return(1);}
      if (tokens>1) sscanf(insp[1],"%f", &r3);
      if (tokens<3) {r1 = hst[hp].xsize; r2 = hst[hp].ysize; }
      else { if (tokens > 2) sscanf(insp[2], "%f", &r1); 
	     if (tokens > 3) sscanf(insp[3], "%f", &r2); else r2 = r1;
	  }
      r=h_const((int)r1,(int)r2,(D)r3);    /* create a new constant HF */
    } else if (keymatch(insp[0], "yslope",3)) {
      r1 = 1; r2 = 1; r3 = 1;
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
      if (tokens > 2) sscanf(insp[2], "%f", &r2);
      if (tokens > 3) sscanf(insp[3], "%f", &r3);
      r=yslope((D)r1,(D)r2,(D)r3);
    } else if (keymatch(insp[0], "zedge",3)) {
      r1 = 0.5; r2 = 1.0;
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
      if (tokens > 2) sscanf(insp[2], "%f", &r2);
      r=h_zedge((D)r1,(D)r2);
    } else if (keymatch(insp[0], "rotate",4)) {
      r1 = 90;
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
       else printf("Rotating clockwise by 90 degrees.\n");
      r=rotate((int)r1);
    } else if (keymatch(insp[0], "gauss",3)) {
      r1 = 0.5; r2 = 0.5; r3 = 0.3; r4 = 0.5;
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
      if (tokens > 2) sscanf(insp[2], "%f", &r2);
      if (tokens > 3) sscanf(insp[3], "%f", &r3);
      if (tokens > 4) sscanf(insp[4], "%f", &r4);
      r=h_gauss((D)r1,(D)r2,(D)r3,(D)r4);
    } else if (keymatch(insp[0], "ring",4)) {
      r1 = 0.5; r2 = 0.5; r3 = 0.25; r4 = 0.20; r5 = 0.5;
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
      if (tokens > 2) sscanf(insp[2], "%f", &r2);
      if (tokens > 3) sscanf(insp[3], "%f", &r3);
      if (tokens > 4) sscanf(insp[4], "%f", &r4);
      if (tokens > 5) sscanf(insp[5], "%f", &r5);
      r=h_ring((D)r1,(D)r2,(D)r3,(D)r4,(D)r5);
    } else if (keymatch(insp[0], "crater",2)) {
      r1 = 100; r2 = 1; r3 = 1; r4 = 10;
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
      if (tokens > 2) sscanf(insp[2], "%f", &r2);
      if (tokens > 3) sscanf(insp[3], "%f", &r3);
      if (tokens > 3) sscanf(insp[4], "%f", &r4);
      r=h_crater((int)r1,(D)r2,(D)r3,(D)r4);
    } else if (keymatch(insp[0], "cfillarray",3)) {
      r3 = 2.1;
      if ((tokens==1) && (hst[hp].a==NULL)) 
	{printf("must specify a dimension.\n"); return(1);}
      if (tokens==1) {r1 = hst[hp].xsize; r2 = hst[hp].ysize; }
      else { if (tokens > 1) sscanf(insp[1], "%f", &r1); 
	     if (tokens > 2) sscanf(insp[2], "%f", &r2); else r2 = r1;
	     if (tokens > 3) sscanf(insp[3], "%f", &r3);
	  }
      r=fillarray((int) r1,(int) r2, 3.0-r3);
    } else if (keymatch(insp[0], "ffhp",4) ||
	       keymatch(insp[0], "fflp",4) ||
	       keymatch(insp[0], "ffbp",4) ||
	       keymatch(insp[0], "ffbr",4)   ) {
      if (tokens<3)
	{ printf("Error: %s needs two arguments.\n",insp[0]);
	  return(1);}
      sscanf(insp[1], "%f", &r1);
      sscanf(insp[2], "%f", &r2);
      r=h_fourfilt((D)r1, (D)r2, insp[0]);  /* filter real & imag. matrix */
    } else if (keymatch(insp[0], "hpf",3) ||
	       keymatch(insp[0], "lpf",3) ||
	       keymatch(insp[0], "bpf",3) ||
	       keymatch(insp[0], "brf",3)   ) {
      if (tokens>1) sscanf(insp[1], "%f", &r1); else r1=-1;
      if (tokens>2) sscanf(insp[2], "%f", &r2); else r2=0;
      r=h_realfilt((D)r1, (D)r2, insp[0]);  /* filter real matrix */
    } else if (keymatch(insp[0], "fillbasin", 3)) {
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1 = 1;
      if (tokens > 2) sscanf(insp[2],"%f",&r2); else r2 = 1;
      r=h_fillb((int)r1,(D)r2);  
    } else if (keymatch(insp[0], "flowmap", 4)) {
      r= h_find_ua();                /* find uphill area */
    } else if (keymatch(insp[0], "gforge", 2)) {
      if ((tokens==1) && (hst[hp].a==NULL)) r1 = 200;
      else if (tokens==1) r1 = hst[hp].xsize;
      if (tokens > 1) sscanf(insp[1],"%f",&r1);
      if (tokens > 2) sscanf(insp[2],"%f",&r2); else r2 = 2.1;
      r=gforge((int)r1,(float)r2);
    } else if (keymatch(insp[0], "fft", 3)) {
      r=h_fft(1,-2.0);
    } else if (keymatch(insp[0], "ifft", 3)) {
      r=h_fft(-1,-2.0);
    } else if (keymatch(insp[0], "smooth", 2)) {
      if (tokens > 1) sscanf(insp[1], "%f", &r1);
	else r1 = 0.5;
      r=smooth((D)r1);
    } else if  (keymatch(insp[0], "add", 3) ||
		keymatch(insp[0], "sub", 3) ||
		keymatch(insp[0], "mul", 3) ||
		keymatch(insp[0], "div", 3) ||
		keymatch(insp[0], "exp", 3) ||
		keymatch(insp[0], "rmag", 4) ||
		keymatch(insp[0], "comneg", 4) ||
		keymatch(insp[0], "compos", 3)  ) {  
       r1 = 0.0; r2 = 0.0;
       if (tokens > 1) sscanf(insp[1], "%f", &r1);
       if (tokens > 2) sscanf(insp[2], "%f", &r2);
       r=composit(insp[0],(int)r1,(int)r2);
    } else if  (
		keymatch(insp[0], "+", 1) ||
		keymatch(insp[0], "-", 1) ||
		keymatch(insp[0], "*", 1) ||
		keymatch(insp[0], "/", 1) ||
		keymatch(insp[0], "^", 1) ||
		keymatch(insp[0], "invert", 3) ||
		keymatch(insp[0], "pow", 3) ||
		keymatch(insp[0], "abs", 3) ||
		keymatch(insp[0], "sin", 3) ||
		keymatch(insp[0], "log", 3) ||
		keymatch(insp[0], "asin", 4) ||
		keymatch(insp[0], "tan",  3) ||
		keymatch(insp[0], "atan", 4) ||    
		keymatch(insp[0], "modulus", 3) ||    
		keymatch(insp[0], "discrete", 3) ||    
		keymatch(insp[0], "floor", 4) ||    
		keymatch(insp[0], "ceiling", 4)     ) {  
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1 = 1.0;
      if (tokens > 2) sscanf(insp[2],"%f",&r2); else r2 = 0.0;
      r = oneop(insp[0],(D)r1,(D)r2);
    } else if (keymatch(insp[0], "diff", 4) || keymatch(insp[0],"dif2",4)) {
      r = h_diff(insp[0]);
    } else if (keymatch(insp[0],"lslope",4) || keymatch(insp[0],"lcurve",4)) {
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1 = 0.01;
      if (tokens > 2) sscanf(insp[2],"%f",&r2); else r2 = 1;
      r=h_slopelim(insp[0],(D)r1,(int)r2); /* slope-dependent smoothing */
    } else if (keymatch(insp[0], "nsmooth", 3)) {
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1 = 1;
      if (tokens > 2) sscanf(insp[2],"%f",&r2); else r2 = FLT_MIN;
      if (tokens > 3) sscanf(insp[3],"%f",&r3); else r3 = FLT_MAX;
      r=h_nsmooth((int) r1,(D)r2,(D)r3);  /* elevation-dependent smoothing */
    } else if (keymatch(insp[0], "list", 1)) {
      if (tokens < 2) {   list_stack();
      } else if (keymatch(insp[1], "stack", 1)) {     list_stack();
      } else if (keymatch(insp[1], "buffers", 1)) {   list_stack();
      } else if (keymatch(insp[1], "variables", 1)) { list_vars();
      } else if (keymatch(insp[1], "formats", 1))   { pr_help("formats");
      } else { pr_help("list");
      }
    } else if (keymatch(insp[0], "last", 2)) {
      if (tokens < 2) r1 = 10; else sscanf(insp[1], "%f", &r1);
      r=show_hist((int)r1);
    } else if (keymatch(insp[0], "set", 3)) {  /* set / show internal variables */
      if (tokens < 3) show_vars();
      else r=set_vars(insp[1],insp[2]);
    } else if (keymatch(insp[0], "seed", 2)) {
      if (tokens < 2) show_vars();
      else r=set_vars("seed",insp[1]);
    } else if (keymatch(insp[0], "ls", 2) || keymatch(insp[0],"dir",3)) {
      if (tokens < 2) listdir(".");
      else {strcpy(buf,insp[1]); listdir(buf); }
    } else if (keymatch(insp[0], "gls", 3) || keymatch(insp[0],"gdir",3)) {
      if (tokens < 2) glistdir(".");
      else {strcpy(buf,insp[1]); glistdir(buf); }
    } else if (keymatch(insp[0], "load", 2)) {
      if (tokens < 2) printf("must specify a filename\n");
      else r=loadhf(insp[1]);
    } else if (keymatch(insp[0], "save", 2)) {
      if (tokens < 2) printf("must specify a filename\n");
      else r=savehf(insp[1]);
    } else if (keymatch(insp[0], "run", 3)) {
      if (takefile) { 
	printf("Error: nested command files not supported.\n");
	exit(1);
      } else if ((cmd_stream = fopen(insp[1],"r")) == NULL) {
	  printf("Error: unable to open cmd file %s\n",insp[1]);
	  cmd_stream = stdin; r=1;
      } else {
	printf("Reading command file %s\n",insp[1]);
	takefile = TRUE;
      }
    } else if (keymatch(insp[0], "rm", 2) || keymatch(insp[0],"del",3)) {
      if (tokens < 2) printf("must specify a filename to delete\n");
      else r=f_delete(insp[1]);
    } else if (keymatch(insp[0], "startlog", 4)) {
      if (tokens < 2) strcpy(logname,"log-hf.txt");
       else strcpy(logname,insp[1]);
      r=startlog(logname);
    } else if (keymatch(insp[0], "stoplog", 4)) {
       r=stoplog();
    } else if (keymatch(insp[0], "view3d", 1)) {
	if (view_current) {
	  printf("Already in view mode. Use 'v' to refresh.\n");
	  r=1;
	} else r=show_grid(hst[hp],insp,tokens);
    } else if (keymatch(insp[0], "tshow", 3)) {
       r1 = 78; r2 = 22;  /* text mode 80x24 */
       if (tokens > 1) sscanf(insp[1], "%f", &r1);
       if (tokens > 2) sscanf(insp[2], "%f", &r2);
       r=h_tshow((int)r1,(int)r2);
    } else if (keymatch(insp[0], "show", 1)) {    
	r=show_flat(hst[hp],insp,tokens);
    } else if (keymatch(insp[0], "equalize", 2)) {
       r1 = 1.0; r2 = 1000;  /* equalization fraction and bins */
       if (tokens > 1) sscanf(insp[1], "%f", &r1);
       r=histeq(r1);
    } else if (keymatch(insp[0], "histogram", 2)) {
       if (tokens > 1) sscanf(insp[1], "%f", &r1); else r1=20;
       r=h_hist((int)r1);
    } else if (keymatch(insp[0], "hshift", 3)) {
       if (tokens > 1) sscanf(insp[1], "%f", &r1); else r1=0.0;
       r=h_hshift(r1);
    } else if (keymatch(insp[0], "normalize", 3)) {
       r1 = 0.0; r2 = 1.0;  /* normalize to range [0..1] */
       if (tokens > 1) sscanf(insp[1], "%f", &r1);
       if (tokens > 2) sscanf(insp[2], "%f", &r2);
       r=norm(r1, r2);
    } else if (keymatch(insp[0], "nrm2", 3)) {
       r1 = 0.0; r2 = 1.0;  /* normalize to max val 1.0 */
       if (tokens > 1) sscanf(insp[1], "%f", &r1);
	else r1=1.0;
       r=nrm2(r1);
    } else if (keymatch(insp[0], "peak", 2)) {
       r1 = 0.5; r2 = 0.5;
       if (tokens > 1) sscanf(insp[1], "%f", &r1);
       if (tokens > 2) sscanf(insp[2], "%f", &r2);
       r=peak((D)r1, (D)r2);
    } else if (keymatch(insp[0], "tile?", 5)) {  r=is_tilable(1);
    } else if (keymatch(insp[0], "negate", 3)) { r=negate();
    } else if (keymatch(insp[0], "pop", 3)) {    r=hpop();
    } else if (keymatch(insp[0], "dup", 3)) {    r=hdup();
    } else if (keymatch(insp[0], "swap", 3)) {   r=hswap();
    } else if (keymatch(insp[0], "cswap", 2)) {   r=c_swap();
    } else if (keymatch(insp[0], "cjoin", 2)) {   r=c_join();
    } else if (keymatch(insp[0], "csplit", 3)) {   r=c_split();
    } else if (keymatch(insp[0], "cdiff", 4)) {   r=c_diff();
    } else if (keymatch(insp[0], "cint", 4)) {   r=c_int();
    } else if (keymatch(insp[0], "cmag", 4)) {   r=c_mag();
    } else if (keymatch(insp[0], "cpolar", 4)) {   r=c_convert(0);
    } else if (keymatch(insp[0], "crect", 4)) {   r=c_convert(1);
    } else if (keymatch(insp[0], "rot3", 4)) {   r=hrot3();
    } else if (keymatch(insp[0], "rot", 3))  { 
      if (tokens > 1) {
	printf("Error: ROT takes no arguments. Did you mean ROTATE?\n");
	r=1;}
      else r=hrot();
    } else if (keymatch(insp[0], "joinx", 5)) {   r=h_join(1);
    } else if (keymatch(insp[0], "joiny", 5)) {   r=h_join(0);
    } else if (keymatch(insp[0], "half", 3)) {   r=h_half();
    } else if (keymatch(insp[0], "double", 4)) {  
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1=0.5;
      if (tokens > 2) sscanf(insp[2],"%f",&r2); else r2=0.0;
      r=h_double((D)r1,(D)r2);
    } else if (keymatch(insp[0], "twist", 4)) {  
      r2=0.5; r3=0.5; r4=0; r5=0;
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1=1.0;
      if (tokens > 2) {
	sscanf(insp[2],"%f",&r2);
	if (tokens > 3) sscanf(insp[3],"%f",&r3); else r3=r2;
      }
      if (tokens > 4) sscanf(insp[4],"%f",&r4);
      if (tokens > 5) sscanf(insp[5],"%f",&r5);
      r=h_warp(0,(D)r2,(D)r3,(D)r1,(int)r4,(int)r5);
    } else if (keymatch(insp[0], "cwarp", 4)) {
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1=1.0;
      r=h_cwarp(0, (D)r1);  /* warp with complex control */
    } else if (keymatch(insp[0], "bloom", 4)) {
      r2=0.5; r3=0.5; r4 = 0; r5=0;
      if (tokens > 1) sscanf(insp[1],"%f",&r1); else r1=1.0;
      if (tokens > 2) {
	sscanf(insp[2],"%f",&r2);
	if (tokens > 3) sscanf(insp[3],"%f",&r3); else r3=r2;
      }
      if (tokens > 4) sscanf(insp[4],"%f",&r4);
      if (tokens > 5) sscanf(insp[5],"%f",&r5);
      r=h_warp(1,(D)r2,(D)r3,(D)r1,(int)r4,(int)r5);
    } else if (keymatch(insp[0], "clear", 5))  { r=clear_stack(stacksize);
    } else if (keymatch(insp[0], "name", 3)) {
      if (tokens < 2) printf("must specify a hf name.\n");
      else r=hname(insp[1]);
    } else if (keymatch(insp[0], "author", 2)) { pr_help("author");
    } else if (keymatch(insp[0], "warranty", 3)) { pr_help("warranty");
    } else if (keymatch(insp[0], "cd..", 4)) {
      if (chdir("..")) perror(insp[0]);
    } else if (keymatch(insp[0], "cd", 2)) {
      if (chdir(insp[1])) perror(insp[1]);
    } else if (keymatch(insp[0], "sleep", 5) || 
	       keymatch(insp[0], "wait", 1))    { 
       if (tokens > 1) sscanf(insp[1], "%f", &r1); else r1=5;
       printf("paused for %d seconds",(U)r1);
       fflush(stdout);
       sleep((U)r1);
       printf("\n");
    } else if (keymatch(insp[0], "quit", 4)) {   
       if (takefile) {
	 fclose(cmd_stream); 
	 cmd_stream=stdin;
	 takefile = FALSE;
	 if (startint) return(1);
       }
       cleanup();
       int_done = TRUE; 
    } else if (keymatch(insp[0], "qui", 1)) {   
      printf("Type 'quit' to exit HF-Lab.\n\n");
      return(1);
    } else if (keymatch(insp[0], "version", 3)) {   
      printf("The version number is:  %s\n",VERSION_NUMBER);
    } else {
      printf("'%s' not understood.\n",insp[0]);
      return(1);
    }  /* end if-else keymatch() */
  return(r); 
} /* h_interp() */

/* cmd_seq(): take a string of commands sep. by ';' in the global cmd_instr[]
 * starting at offset startp  (ie, cmd_instr[startp] )
 * parse into tokens starting at character <startp>, and then
 * call the command executor.
 */

/* cmd_seq(int ptr)  run command sequence a;b;c from string cmd_instr[]   
 *                 starting from position in string indexed by ptr
 */

int cmd_seq(int cmd_ptr)  /* run command sequence a;b;c from cmd_instr[] */
{
 char oneline[CLINELEN];        /* a single command */
 char insp[MAXTOK][TOKLEN];     /* tokens array, each TOKLEN-1 long max. */
 int tokens;                    /* number of tokens on this line */
 int i,j,coff;
 int reps;                      /* number of times to repeat following cmds */
 char onech;                    /* a single character of cmd line */
 int result;                    /* return value of interp() */


   /* first task: if command starts "rep n [...]" then enter a do-loop
      to call cmd_seq() again n times, with the remaining commands "[...]"
    */

   while (cmd_instr[cmd_ptr]==SEPTOR) cmd_ptr++;  /* skip over leading SEPTOR */
   if (cmd_instr[cmd_ptr]=='#') return(0);         /* done if reached comment marker */

   coff = cmd_ptr;   /* start at this character position in input line */
   parse_1tok(cmd_instr, &coff, oneline); /* returns 1st token in oneline[] */

   if (keymatch(oneline, "rep", 3)) {     /* if REPEAT command has been found */
/*     printf("found repeat:%s at position %d in %s\n",oneline,coff,cmd_instr); */
     parse_1tok(cmd_instr, &coff, oneline); /* get repeat count */
     reps = atoi(oneline);                  /* as integer */
/*     printf("# of reps = %d\n",reps); */
     result = FALSE;  				/* assume no error to start */
     for (i=0;i<reps && result==FALSE;i++) {
       result=cmd_seq(coff);               /* recursive call to execute loop */
      /* printf("result from loop %d is %d.\n",i,result); */
     }
     return(result);  
   } else {                             /* no REPEAT */

     coff = cmd_ptr;  /* no REP: reset command string pointer to start */

     i = -1;   /* parse input into separate commands, then separate tokens */
     do {      /* get a single command into oneline[] from many in instr[] */
       i++;
       onech = cmd_instr[i+coff];
       oneline[i] = onech;
     } while (oneline[i]!=SEPTOR && onech!='\0');  /* SEPTOR separates commands ; */
	
/*     printf ("offset=%d\n",i); */

	/* here, have arrived at the end of a command (SEPTOR) or a line ('\0') */
     if (oneline[i]==SEPTOR) {  /* command separator found */
	oneline[i]='\0';        /* replace separator with null termination */
        parse_line(oneline,&tokens,insp);  /* parse oneline[] into insp[] */
        result = h_interp(tokens,insp);    /* execute command in insp[] */
        for (j=0;j<tokens;j++) insp[j][0] = '\0'; /* erase the old tokens (?) */
	if (result!=0) return(result);
        else return(cmd_seq(i+coff));      /* recursive call for rest of cmds */

     } else {  /* end of input line found */
        parse_line(oneline,&tokens,insp);  /* parse oneline[] into insp[] */
        return( h_interp(tokens,insp));    /* execute command in insp[] */
        for (j=0;j<tokens;j++) insp[j][0] = '\0'; /* erase the old tokens (?) */
     }

   } /* end if REPEAT */

} /* cmd_seq() */


/* cmd_seq1()  this is the older version of cmd_seq(); slated for demolition */

int cmd_seq1(char instr[CLINELEN])  /* run a command sequence a;b;c etc */
{
 char oneline[CLINELEN];        /* a single command */
 char insp[MAXTOK][TOKLEN];     /* tokens array, each TOKLEN-1 long max. */
 int tokens;                    /* number of tokens on this line */
 int i,coff;
 char onech;                    /* a single character of cmd line */
 int result;                    /* return value of interp() */

   coff = 0;
   do {      /* until command returns error or end of instr[] */

    i = -1;   /* parse input into separate commands, then separate tokens */
    do {      /* get a single command into oneline[] from many in instr[] */
      i++;
      onech = instr[i+coff];
      oneline[i] = onech;
    } while (oneline[i]!=SEPTOR && onech!='\0');
    if (oneline[i]==SEPTOR) oneline[i]='\0';
    coff += i+1;
    
    parse_line(oneline,&tokens,insp);  /* parse oneline[] into insp[] */
    /*    printf("coff = %d  read %d tokens %s\n",coff,tokens,insp[0]); */

    result = h_interp(tokens,insp);    /* execute command in insp[] */
    for (i=0;i<tokens;i++) insp[i][0] = '\0';  /* erase the old tokens */
   
   } while (onech!='\0' && result==0);

 return(result);
} /* cmd_seq1() */

/* -------------- parse input string *instr into strings -------------------*/

int parse_line(char *instr, int *t_ptr, char insp[MAXTOK][TOKLEN])
{
 int i,j,start,tlen;

    i = 0;                      /* position pointer in input string */
    j = 0;
    for (;instr[i]==' ';i++);   /* remove leading spaces */
    /* if (i>0) printf("%d leading spaces ignored.\n",i); */
    do {
	start = i;                             /* start of the j-th token */
	for (;instr[i]!=' '&&instr[i]!='\0';i++); /* scan over non-spaces */
	tlen = i-start;
	if (tlen >= TOKLEN) {
	  printf("String too long: %s\n",&instr[start]);
	  tlen = TOKLEN;
	}
	strncpy(insp[j],&instr[start],tlen);   /* store this token */
	insp[j][i-start] = '\0';                  /* null-terminate string */
	/* printf("storing #%d: %s\n",j,insp[j]); */
	if (++j > MAXTOK) {
	  printf("Input line too long!\n");
	  break;
	}
	if (instr[i]=='\0') break;      /* exit if end of instr reached */
	for (;instr[i]==' ';i++);                 /* scan over spaces */
    } while (instr[i] != '\0');           /* stop at end of input string */
   
   /*  -- input parsed into TOKENS strings stored in insp[0..tokens-1] -- */
    t_ptr[0] = j;                        /* total number of tokens found */
    return(j);
}   /* parse_line() */

/* -------------- parse one token from a string -------------------*/
/* input string in instr[], pointer in t_ptr, output token in tok_s */ 
/* will copy up to a maximum of TOKLEN characters into tok_s[]      */
/* token ends with space, SEPTOR, or end of string '\0\             */
/* increments t_ptr to point to final position in string            */

int parse_1tok(char *instr, int *t_ptr, char *tok_s)
{
 int i,start,tlen;

    i = *t_ptr;      /* position pointer in input string */
    for (;instr[i]==' ';i++);   /* remove leading spaces */

    /* if (i>0) printf("%d leading spaces ignored.\n",i); */
	
	start = i;                             /* start of this token */

	for (;instr[i]!=' '&&instr[i]!='\0'&&instr[i]!=SEPTOR;i++); 
					       /* accept non-spaces */
	tlen = i-start;
	if (tlen >= TOKLEN) {
	  printf("String too long: %s\n",&instr[start]);
	  tlen = TOKLEN;
	}
	strncpy(tok_s,&instr[start],tlen);   /* store this token */
	tok_s[i-start] = '\0';                  /* null-terminate string */
	/* printf("storing #%d: %s\n",j,tok_s); */

    *t_ptr = i;
    return(0);
}   /* parse_1tok() */

int cmd_rep(int tokens, char insp[MAXTOK][TOKLEN]) /* repeat commands */
{
 /*
 int i, reps, retval;
 char outsp[MAXTOK][TOKLEN];
      
 if (tokens > 1) sscanf(insp[1],"%d",&reps); 
 else {
   printf("repeat error: must specify # of repetitions\n");
   return(1);
 }
 printf("repeating %d times.\n",reps);

 for (i=2;i<tokens;i++) {
  strcpy(outsp[i-2],insp[i]);
 }

 do {
   retval = h_interp(
 */
 return(0);
} /* cmd_rep() */
