/* png_io.c  --  Test file to read PNG image. Uses PNGLIB 0.88 and ZLIB 1.0.
 *               Reads non-interlaced images ok, but not interlaced ones.
 *               5/4/96  John P. Beale
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <png.h>

#define U unsigned int

/* png_read() -- read a png file into a heightfield structure
 *               based on example.c from libpng 0.8
 */

int png_read(char *file_name)
{
  int x, y;
  int xsize, ysize;     /* necessary for EL() macro */
  FILE *fp;
  size_t imem;
  int ct;               /* PNG color type */
  png_struct *png_ptr;
  png_info *info_ptr;
  png_byte **rowptr;   /* array of row vectors for image storage */

   /* open the file */
   printf("opening PNG file %s\n",file_name);
   fp = fopen(file_name, "rb");
   if (!fp) {
      printf("Error opening PNG file %s.\n",file_name);
      return(1);
   }

   /* allocate the necessary structures */
   png_ptr = malloc(sizeof (png_struct));
   if (!png_ptr)
   {
      fclose(fp);
      printf("Error opening PNG file %s: out of memory.\n",file_name);
      return(1);
   }

   info_ptr = malloc(sizeof (png_info));
   if (!info_ptr)
   {
      fclose(fp);
      free(png_ptr);
      printf("Error opening PNG file %s: out of memory.\n",file_name);
      return(1);
   }

   /* set error handling */
   if (setjmp(png_ptr->jmpbuf))
   {
      png_read_destroy(png_ptr, info_ptr, (png_info *)0);
      fclose(fp);
      free(png_ptr);
      free(info_ptr);
      /* If we get here, we had a problem reading the file */
      printf("Error reading PNG file %s.\n",file_name);
      return(1);
   }

   /* initialize the structures, info first for error handling */
   png_info_init(info_ptr);
   png_read_init(png_ptr);

   /* set up the input control */
   png_init_io(png_ptr, fp);

   /* read the file information */
   png_read_info(png_ptr, info_ptr);
   ct = info_ptr->color_type;
   xsize = info_ptr->width;
   ysize = info_ptr->height;
   printf("[%d x %d]\n",xsize,ysize);

    /* expand paletted colors into true rgb */
   printf("PNG bit depth is %d.\n",info_ptr->bit_depth);
   if (ct == PNG_COLOR_TYPE_PALETTE)
    printf("PNG file is palette type.\n");
  /*
      png_set_expand(png_ptr);
  */

   /* expand sub-8-bit grayscale images to the full 8 bits */
 /*
   if ((ct == PNG_COLOR_TYPE_GRAY) && (info_ptr->bit_depth < 8))
      png_set_expand(png_ptr);
   */
   /* tell libpng to strip 16 bit depth files down to 8 bits */
   if ( (info_ptr->bit_depth == 16) &&
	(ct != PNG_COLOR_TYPE_GRAY) )
				    /* 16 bit "super-trucolor" image */
	png_set_strip_16(png_ptr);

    png_read_update_info(png_ptr,info_ptr);
    printf("Rowbytes needed: %ld\n",(long)info_ptr->rowbytes);

    if ((ct != PNG_COLOR_TYPE_GRAY) && (ct != PNG_COLOR_TYPE_RGB) &&
	(ct != PNG_COLOR_TYPE_PALETTE) ) {
      printf("Sorry, unsupported PNG file type.\n");
    } else
      {

   if (info_ptr->valid & PNG_INFO_sBIT)
     printf("num signficant bits: %d\n",(int)info_ptr->sig_bit.gray);
   imem = (size_t) xsize;
   if ((info_ptr->bit_depth==16) && (ct == PNG_COLOR_TYPE_GRAY))
    imem *= 2;   /* 16-bit GRAY HF */
   else if (ct == PNG_COLOR_TYPE_RGB) 
    imem *= 3;   /* 24-bit color */
  /* else 
    imem *= 3;
  */ 
  printf("allocating %ld for %d x %d HF %s...\n",(long)imem,
	     xsize,ysize,file_name);
  
  if ((rowptr = (png_byte **) malloc((size_t)ysize*sizeof(png_byte *)) ) == NULL) {
	     fclose(fp); printf("Memory allocation error\n");
	     free(png_ptr); free(info_ptr);
	     return(1);
   } 
   printf("allocating rows...\n");
   for (y=0;y<ysize;y++) {
/*     printf("%d",y);fflush(stdout); */
     if ((rowptr[y] = (png_byte *)malloc(imem)) == NULL) {
	     fclose(fp); printf("Memory allocation error\n");
	     for (x=0;x<y;x++) free(rowptr[x]);
	     free(rowptr);
	     free(png_ptr); free(info_ptr);
	     return(1);
     } 
   }
   printf("allocated %ld for %d x %d HF %s...reading image.\n",(long)imem,
	     xsize,ysize,file_name);

   png_read_image(png_ptr,rowptr);
   printf("PNG Image read sucessfully!\n");

   /* clean up after the read, and free any memory allocated */
   png_read_destroy(png_ptr, info_ptr, (png_info *)0);

   /* free the structures */
   for (y=0;y<ysize;y++) free(rowptr[y]);
   free(rowptr);
   free(png_ptr);
   free(info_ptr);
      }

   fclose(fp);
   return(0);

} /* png_read */


/* a simple main() driver to read in one png file */

void main(int argc, char *argv[])
{
 if (argc > 1)
  png_read(argv[1]);
 else png_read("a.png");
}


